
from typing import TypeVar, Generic, List, Union
from picsellia import exceptions as exceptions
from picsellia.sdk.dao import Dao
from picsellia.sdk.connexion import Connexion
from beartype import beartype

from picsellia.bcolors import bcolors
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

T = TypeVar('T', bound=Dao)

class MultiObject(Generic[T]):

    @beartype
    def __init__(self, connexion: Connexion, items: List[T]) -> None:
        self._connexion = connexion

        if items == []:
            raise exceptions.NoDataError("A MultiObject can't be empty")

        self.items : list[T] = items
    
    @property
    def connexion(self) -> Connexion:
        return self._connexion

    @beartype
    def __str__(self,) -> str:
        return "{}MultiObject{} object, size: {}".format(bcolors.GREEN, bcolors.ENDC, len(self))

    @beartype
    def __getitem__(self, key) -> None:
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __add__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __sub__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __mul__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __truediv__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __floordiv__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __mod__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __pow__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __rshift__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __lshift__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __and__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __or__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __xor__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __lt__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __gt__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __le__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __ge__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")

    @beartype
    def __eq__(self, other):
        if isinstance(other, List):
            if len(other) != len(self.items):
                return False

            for item in self.items:
                if item not in other:
                    return False

        if isinstance(other, MultiObject):
            if self.connexion != other.connexion or len(self.items) != len(other.items):
                return False

            for item in self.items:
                if item not in other.items:
                    return False
        else:
            return False

        return True

    @beartype
    def __iadd__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")
    
    @beartype
    def __isub__(self, other):
        raise NotImplementedError("Not implementend. Please contact support")


    @beartype
    def __len__(self) -> int:
        return len(self.items)

    @beartype
    def assert_enough_items(self) -> None:
        if len(self.items) < 2:
            raise exceptions.NoDataError("You can't remove from this list the last item. A MultiObject can't be empty")

    @beartype
    def assert_same_connexion(self, other: Union['MultiObject', Dao]) -> None:
        if self.connexion != other.connexion:
            raise exceptions.PicselliaError('Objects that you manipulate does not come from the same client.')

    @beartype
    def __delitem__(self, i: int) -> None:
        self.assert_enough_items()
        del self.items[i]

    @beartype
    def __setitem__(self, i: int, v: T) -> None:
        self.assert_same_connexion(v)
        self.items[i] = v

    @beartype
    def append(self, x: T) -> None:
        self.assert_same_connexion(x)
        self.items.append(x)

    @beartype
    def extend(self, objects: Union[T, 'MultiObject[T]', List[T]]) -> None:
        if isinstance(objects, Dao):
            self.assert_same_connexion(objects)
        elif isinstance(objects, MultiObject) or isinstance(objects, List):
            for x in objects:
                self.assert_same_connexion(x)
        else:
            raise TypeError(objects)
        self.items.extend(objects)

    @beartype
    def insert(self, i: int, x: T) -> None:
        self.assert_same_connexion(x)
        self.items.insert(i, x)

    @beartype
    def remove(self, x: T) -> None:
        self.assert_enough_items()
        self.items.remove(x)

    @beartype
    def pop(self, i: int = None) -> T:
        self.assert_enough_items()
        if i:
            return self.items.pop(i)
        else:
            return self.items.pop()

    @beartype
    def clear(self) -> None:    
        raise exceptions.NoDataError("You can't clear a MultiObject.")

    @beartype
    def index(self, x: T, start: int = None, end: int = None) -> int:
        if start:
            if end:
                return self.items.index(x, start, end)
            else:
                return self.items.index(x, start)
        else:
            return self.items.index(x)

    @beartype
    def count(self, x: T) -> int:
        return self.items.count(x)

    @beartype
    def sort(self, key = None, reverse = False) -> None:    
        self.items.sort(key=key, reverse=reverse)

    @beartype
    def reverse(self) -> None:    
        self.items.reverse()

    @beartype
    def copy(self) -> 'MultiObject':
        raise NotImplementedError("Not implementend. Please contact support")





import logging
from beartype import beartype
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connexion
from picsellia.sdk.dao import Dao
from picsellia.sdk.run import Run
from picsellia.bcolors import bcolors
import warnings
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
warnings.filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)
from picsellia.types.schemas import ScanSchema

logger = logging.getLogger('picsellia')

class Scan(Dao):

    def __init__(self, connexion: Connexion, data: dict) -> None:
        scan = ScanSchema(**data)
        super().__init__(connexion, scan.id)
        self._name = scan.name

    @property
    def name(self) -> str:
        return self._name

    def __str__(self,): return "{}Scan '{}' {} (id: {})".format(bcolors.BLUE, self.name, bcolors.ENDC, self.id)

    @exception_handler
    @beartype
    def launch(self,) -> None:
        """Distribute runs remotely for this scan.

        :information-source: The remote environment has to be setup prior launching the experiment.
        It defaults to our remote training engine.

        Examples:
            ```python
                scan.launch()
            ```
        """
        self.connexion.post('/sdk/v1/scan/{}/launch'.format(self.id))

    @exception_handler
    @beartype
    def get_run_by_id(self, id: str) -> Run:
        """Retrieve a run object by its id.

        Examples:
            ```python
                scan.get_run_by_id("cb750009-4e09-42bb-8c84-cc78aa004bf0")
            ```
        Arguments:
            id (str): id (primary key) of the run on Picsellia

        Returns:
            A (Run) object manipulable
        """
        r = self.connexion.get('/sdk/v1/run/{}'.format(id)).json()
        return Run(self.connexion, r["run"])

    @exception_handler
    @beartype
    def get_next_run(self,) -> Run:
        """Get next available Run for Scan.

        Examples:
            ```python
                scan.get_next_run()
            ```

        Returns:
            A (Run) object manipulable
        """
        r = self.connexion.get('/sdk/v1/scan/{}/run/next'.format(self.id)).json()
        return Run(self.connexion, r["run"])

    @exception_handler
    @beartype
    def delete(self,) -> None:
        """Delete this scan from the platform.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
                scan.delete()
            ```
        """
        self.connexion.delete('/sdk/v1/scan/{}'.format(self.id))
        logger.info("Scan (id: {}) deleted from platform.".format(self.id))



from picsellia.sdk.connexion import Connexion


class Dao:

    def __init__(self, connexion: Connexion, id: str) -> None:
        self._id = id
        self._connexion = connexion

    @property
    def connexion(self) -> Connexion:
        return self._connexion

    @property
    def id(self) -> Connexion:
        return self._id

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, type(self)):
            return self._id == __o.id and self._connexion == __o.connexion

        return False



from dataclasses import Field
from typing import Dict, List, Optional
from pydantic import BaseModel, Field

class DatalakeSchema(BaseModel):
    id: str = Field(alias='datalake_id')

class OrganizationSchema(BaseModel):
    id: str = Field(alias='organization_id')
    name: str = Field(alias='organization_name')
class DatasetSchema(BaseModel):
    id: str = Field(alias='dataset_id')
    name: str = Field(alias='dataset_name')
    version: str = Field(alias='version')

class ModelSchema(BaseModel):
    id: str = Field(alias='model_id')
    name: str = Field(alias='network_name')
    type: str = Field(alias='type')

class ProjectSchema(BaseModel):
    id: str = Field(alias='project_id')
    name: str = Field(alias='project_name')

class DeploymentSchema(BaseModel):
    id: str = Field(alias='pk')
    name: str = Field(alias='name')
    oracle_host: Optional[str] = Field(default=None, alias='oracle_url')
    serving_host: Optional[str] = Field(default=None, alias='serving_url')

class DataSchema(BaseModel):
    picture_id: str = Field(alias='picture_id')
    external_url: str = Field(alias='external_url')
    internal_key: str = Field(alias='internal_key')

class PictureSchema(BaseModel):
    picture_id: str = Field(alias='picture_id')
    external_url: str = Field(alias='external_url')
    internal_key: str = Field(alias='internal_key')
    width: int = Field(alias='width')
    height: int = Field(alias='height')
    tags: Optional[List[str]] = Field(default=None, alias='tag')

class ExperimentSchema(BaseModel):
    id: str = Field(alias='id')
    name: str = Field(alias='name')
    files: Optional[List[dict]] = Field(default=[], alias='files')

class UserSchema(BaseModel):
    username: str = Field(alias='username')

class WorkerSchema(BaseModel):
    id: str = Field(alias='pk')
    user: UserSchema = Field(alias='user')

class ScanSchema(BaseModel):
    id: str = Field(alias='id')
    name: str = Field(alias='name')

class RunSchema(BaseModel):
    id: str = Field(alias='id') 
    config: dict = Field(alias='config') 
    requirements: List[dict] = Field(default=[], alias='requirements') 
    script: Optional[str] = Field(default=None, alias='script') 
    script_object_name: Optional[str] = Field(default=None, alias='script_object_name') 



from dataclasses import Field
from typing import List
from pydantic import BaseModel, Field, root_validator, validator

def labels_validator(v: List[int]) -> List[int]:
    if not isinstance(v, List):
        raise TypeError(v)
    
    if len(v) < 1:
        raise ValueError('There must be at least one id')

    return v

def boxes_validator(v: List[List[int]]) -> List[List[int]]:
    if not isinstance(v, List):
        raise TypeError(v)
    
    if len(v) < 1:
        raise ValueError('There must be at least one id')

    return v

def detection_scores_validator(v: List[float]) -> List[float]:
    if not isinstance(v, List):
        raise TypeError(v)
    
    if len(v) < 1:
        raise ValueError('There must be at least one score')

    return v

def masks_validator(v: List[List[int]]) -> List[List[int]]:
    if not isinstance(v, List):
        raise TypeError(v)
    
    if len(v) < 1:
        raise ValueError('There must be at least one id')

    return v
    


class PredictionFormat(BaseModel):
    pass

class ClassificationPredictionFormat(PredictionFormat):
    labels: List[int] = Field(alias="label_id")
    detection_scores: List[float] = Field(alias="score")

    @validator('labels', pre=True)
    def to_label_ids(cls, v):
        if isinstance(v, int):
            return [v]
        elif isinstance(v, List):
            return v
        raise TypeError(v)

    @validator('detection_scores', pre=True)
    def to_score_list(cls, v):
        if isinstance(v, float) or isinstance(v, int):
            return [float(v)]
        elif isinstance(v, List):
            return v
        raise TypeError(v)

    _validate_labels = validator('labels', allow_reuse=True)(labels_validator)
    _validate_scores = validator('detection_scores', allow_reuse=True)(detection_scores_validator)
    
class DetectionPredictionFormat(PredictionFormat):
    labels: List[int] = Field(alias='label_ids')
    boxes: List[List[int]] = Field(alias='boxes')
    detection_scores: List[float] = Field(alias='detection_scores')

    _validate_labels = validator('labels', allow_reuse=True)(labels_validator)
    _validate_scores = validator('detection_scores', allow_reuse=True)(detection_scores_validator)
    _validate_boxes = validator('boxes', allow_reuse=True)(boxes_validator)

    @root_validator
    def check_sizes(cls, values):
        labels, detection_scores, boxes = values.get('labels'), values.get('detection_scores'), values.get('boxes')
        
        if labels is None or detection_scores is None or boxes is None is None or len(labels) != len(detection_scores) or len(boxes) != len(labels):
            raise ValueError('incoherent lists')

        return values


class SegmentationPredictionFormat(PredictionFormat):
    labels: List[int] = Field(alias='label_ids')
    boxes: List[List[int]] = Field(alias='boxes')
    detection_scores: List[float] = Field(alias='detection_scores')
    masks: List[List[int]] = Field(alias="masks")

    _validate_labels = validator('labels', allow_reuse=True)(labels_validator)
    _validate_scores = validator('detection_scores', allow_reuse=True)(detection_scores_validator)
    _validate_boxes = validator('boxes', allow_reuse=True)(boxes_validator)
    _validate_masks = validator('masks', allow_reuse=True)(masks_validator)

    @root_validator
    def check_sizes(cls, values):
        labels, detection_scores, boxes, masks = values.get('labels'), values.get('detection_scores'), values.get('boxes'), values.get('masks')
        
        if labels is None or detection_scores is None or boxes is None or masks is None or len(labels) != len(detection_scores) or len(boxes) != len(labels) or len(masks) != len(labels):
            raise ValueError('incoherent lists')

        return values
